/**************************************************************************

libtbm

Copyright 2012 Samsung Electronics co., Ltd. All Rights Reserved.

Contact: SooChan Lim <sc1.lim@samsung.com>, Sangjin Lee <lsj119@samsung.com>
Boram Park <boram1288.park@samsung.com>, Changyeon Lee <cyeon.lee@samsung.com>

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sub license, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
IN NO EVENT SHALL PRECISION INSIGHT AND/OR ITS SUPPLIERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

**************************************************************************/

#include "config.h"

#include "tbm_bufmgr.h"
#include "tbm_bufmgr_int.h"
#include "tbm_drm_helper.h"
#include "list.h"

#include <sys/resource.h>

int trace_mask = 0;

#ifdef HAVE_DLOG
int bDlog;
#endif

tbm_bufmgr gBufMgr;
int b_dump_queue;

static pthread_mutex_t gLock = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t tbm_bufmgr_lock = PTHREAD_MUTEX_INITIALIZER;
static double scale_factor = 0;
void _tbm_bufmgr_mutex_unlock(void);

//#define TBM_BUFMGR_INIT_TIME

#define MAX_SIZE_N(dest)	(sizeof(dest) - strlen(dest) - 1)

/* check condition */
#define TBM_BUFMGR_RETURN_IF_FAIL(cond) {\
	if (!(cond)) {\
		TBM_ERR("'%s' failed.\n", #cond);\
		_tbm_set_last_result(TBM_ERROR_INVALID_PARAMETER);\
		_tbm_bufmgr_mutex_unlock();\
		return;\
	} \
}

#define TBM_BUFMGR_RETURN_VAL_IF_FAIL(cond, val) {\
	if (!(cond)) {\
		TBM_ERR("'%s' failed.\n", #cond);\
		_tbm_set_last_result(TBM_ERROR_INVALID_PARAMETER);\
		_tbm_bufmgr_mutex_unlock();\
		return val;\
	} \
}

/* LCOV_EXCL_START */

static void
_tbm_bufmgr_check_bo_cnt(tbm_bufmgr bufmgr)
{
	static int last_chk_bo_cnt = 0;

	if ((bufmgr->bo_cnt >= 500) && ((bufmgr->bo_cnt % 20) == 0) &&
		(bufmgr->bo_cnt > last_chk_bo_cnt)) {
		TBM_DBG("============TBM BO CNT DEBUG: bo_cnt=%d\n", bufmgr->bo_cnt);
		tbm_bufmgr_debug_show(bufmgr);
		last_chk_bo_cnt = bufmgr->bo_cnt;
	}
}

static void
_tbm_bufmgr_initialize_bo(tbm_bufmgr bufmgr, tbm_bo bo, int flags)
{
	bo->bufmgr = bufmgr;
	bo->flags = flags;
	bo->magic = TBM_BO_MAGIC;
	bo->ref_cnt = 1;

	LIST_INITHEAD(&bo->user_data_list);

	bufmgr->bo_cnt++;
	LIST_ADD(&bo->item_link, &bufmgr->bo_list);
}

void
_tbm_bufmgr_mutex_lock(void)
{
	pthread_mutex_lock(&tbm_bufmgr_lock);
}

void
_tbm_bufmgr_mutex_unlock(void)
{
	pthread_mutex_unlock(&tbm_bufmgr_lock);
}

static int
_tbm_util_get_max_surface_size(int *w, int *h)
{
	tbm_surface_info_s info;
	tbm_surface_h surface = NULL;
	int count = 0;

	*w = 0;
	*h = 0;

	if (gBufMgr == NULL || LIST_IS_EMPTY(&gBufMgr->surf_list))
		return count;

	LIST_FOR_EACH_ENTRY(surface, &gBufMgr->surf_list, item_link) {
		if (tbm_surface_get_info(surface, &info) == TBM_SURFACE_ERROR_NONE) {
			count++;
			if (*w < info.width)
				*w = info.width;
			if (*h < info.height)
				*h = info.height;
		}
	}

	return count;
}

static void
_tbm_util_get_appname_brief(char *brief)
{
	char delim[] = "/";
	char *token = NULL;
	char temp[255] = {0,};
	char *saveptr = NULL;

	token = strtok_r(brief, delim, &saveptr);

	while (token != NULL) {
		memset(temp, 0x00, 255 * sizeof(char));
		strncpy(temp, token, 254 * sizeof(char));
		token = strtok_r(NULL, delim, &saveptr);
	}

	snprintf(brief, sizeof(temp), "%s", temp);
}

static void
_tbm_util_get_appname_from_pid(long pid, char *str)
{
	char fn_cmdline[255] = {0, }, cmdline[255];
	FILE *fp;
	int len;

	if (pid <= 0) return;

	snprintf(fn_cmdline, sizeof(fn_cmdline), "/proc/%ld/cmdline", pid);

	fp = fopen(fn_cmdline, "r");
	if (fp == 0) {
		TBM_ERR("cannot file open %s\n", fn_cmdline);
		return;
	}

	if (!fgets(cmdline, 255, fp)) {
		TBM_ERR("fail to get appname for pid(%ld)\n", pid);
		fclose(fp);
		return;
	}

	fclose(fp);

	len = strlen(cmdline);
	if (len < 1)
		memset(cmdline, 0x00, 255);
	else
		cmdline[len] = 0;

	snprintf(str, sizeof(cmdline), "%s", cmdline);
}

static void
_tbm_bufmgr_copy_module_data(tbm_bufmgr bufmgr, tbm_module *module)
{
	bufmgr->module_data = module->module_data;
	bufmgr->backend = module->backend;

	bufmgr->backend_module_data = module->backend_module_data;
	bufmgr->bufmgr_data = module->bufmgr_data;
	bufmgr->bufmgr_func = module->bufmgr_func;
	bufmgr->bo_func = module->bo_func;

	bufmgr->use_hal_tbm = module->use_hal_tbm;
	bufmgr->auth_wl_socket_created = module->auth_wl_socket_created;
	bufmgr->auth_fd = module->auth_fd;
	bufmgr->hal_backend = module->hal_backend;
	bufmgr->hal_bufmgr = module->hal_bufmgr;
}

static void
_tbm_bufmgr_reset_modlue_data(tbm_bufmgr bufmgr)
{
	bufmgr->module_data = NULL;
	bufmgr->backend = NULL;

	bufmgr->backend_module_data = NULL;
	bufmgr->bufmgr_data = NULL;
	bufmgr->bufmgr_func = NULL;
	bufmgr->bo_func = NULL;

	bufmgr->use_hal_tbm = 0;
	bufmgr->auth_wl_socket_created = 0;
	bufmgr->auth_fd = -1;
	bufmgr->hal_backend = NULL;
	bufmgr->hal_bufmgr = NULL;
}

/* LCOV_EXCL_STOP */

static tbm_bufmgr
_tbm_bufmgr_init(int fd, int server)
{
	tbm_error_e error = TBM_ERROR_NONE;

#ifdef TBM_BUFMGR_INIT_TIME
	struct timeval start_tv, end_tv;
#endif
	char *env;

#ifdef TBM_BUFMGR_INIT_TIME
	/* get the start tv */
	gettimeofday(&start_tv, NULL);
#endif

	bDlog = 1;

	/* LCOV_EXCL_START */
#ifdef TRACE
	env = getenv("TBM_TRACE");
	if (env) {
		trace_mask = atoi(env);
		TBM_DBG("TBM_TRACE=%s\n", env);
	} else
		trace_mask = 0;
#endif

	pthread_mutex_lock(&gLock);

	_tbm_set_last_result(TBM_ERROR_NONE);

	if (fd >= 0) {
		TBM_WRN("!!!!!WARNING:: The tbm_bufmgr_init DOSE NOT use argument fd ANYMORE.\n");
		TBM_WRN("!!!!!WARNING:: IT WILL BE CHANGED like tbm_bufmgr_init(int fd) --> tbm_bufmgr_init(void).\n");
	}


	/* initialize buffer manager */
	if (gBufMgr) {
		gBufMgr->ref_count++;
		TBM_DBG("reuse  tbm_bufmgr(%p) ref_count(%d) fd(%d)\n", gBufMgr, gBufMgr->ref_count, gBufMgr->fd);
		pthread_mutex_unlock(&gLock);
		return gBufMgr;
	}

	TBM_DBG("bufmgr init\n");

	/* allocate bufmgr */
	gBufMgr = calloc(1, sizeof(struct _tbm_bufmgr));
	if (!gBufMgr) {
		TBM_ERR("error: fail to alloc bufmgr fd(%d)\n", fd);
		_tbm_set_last_result(TBM_ERROR_OUT_OF_MEMORY);
		pthread_mutex_unlock(&gLock);
		return NULL;
	}

	gBufMgr->fd = fd;

	/* set the display_server flag before loading the backend module */
	if (server) {
		TBM_INFO("The tbm_bufmgr(%p) is used by display server. Need to bind the native_display.\n", gBufMgr);
		gBufMgr->display_server = 1;
	}

	/* load bufmgr priv from env */
	gBufMgr->module = tbm_module_load(gBufMgr->fd);
	if (!gBufMgr->module) {
		TBM_ERR("error : Fail to load bufmgr backend\n");
		_tbm_set_last_result(TBM_ERROR_INVALID_OPERATION);
		free(gBufMgr);
		gBufMgr = NULL;
		pthread_mutex_unlock(&gLock);
		return NULL;

	}

	// TODO: this is temporary. it will be removed after finishing refactoring the tbm_module.
	_tbm_bufmgr_copy_module_data(gBufMgr, gBufMgr->module);

	/* check the essential capabilities of tbm_module */
	gBufMgr->capabilities = tbm_module_bufmgr_get_capabilities(gBufMgr->module, &error);
	if (gBufMgr->capabilities == TBM_BUFMGR_CAPABILITY_NONE) {
		TBM_ERR("The capabilities of the backend module is TBM_BUFMGR_CAPABILITY_NONE.");
		TBM_ERR("TBM_BUFMGR_CAPABILITY_SHARE_FD is the essential capability.");
		tbm_module_unload(gBufMgr->module);
		_tbm_set_last_result(error);
		free(gBufMgr);
		gBufMgr = NULL;
		pthread_mutex_unlock(&gLock);
		return NULL;
	}

	if (!(gBufMgr->capabilities & TBM_BUFMGR_CAPABILITY_SHARE_FD)) {
		TBM_ERR("The capabilities of the backend module had no TBM_BUFMGR_CAPABILITY_SHARE_FD.");
		TBM_ERR("The tbm backend has to get TBM_BUFMGR_CAPABILITY_SHARE_FD. ");
		tbm_module_unload(gBufMgr->module);
		_tbm_set_last_result(TBM_ERROR_INVALID_OPERATION);
		free(gBufMgr);
		gBufMgr = NULL;
		pthread_mutex_unlock(&gLock);
		return NULL;
	}

	/* LCOV_EXCL_STOP */

	gBufMgr->ref_count = 1;

	TBM_INFO("create tizen bufmgr:%p ref_count:%d\n",
	    gBufMgr, gBufMgr->ref_count);

	/* setup the bo_lock_type */
	env = getenv("BUFMGR_LOCK_TYPE");
	if (env && !strcmp(env, "always"))
		gBufMgr->bo_lock_type = TBM_BUFMGR_BO_LOCK_TYPE_ALWAYS;
	else if (env && !strcmp(env, "none"))
		gBufMgr->bo_lock_type = TBM_BUFMGR_BO_LOCK_TYPE_NEVER;
	else if (env && !strcmp(env, "once"))
		gBufMgr->bo_lock_type = TBM_BUFMGR_BO_LOCK_TYPE_ONCE;
	else
		gBufMgr->bo_lock_type = TBM_BUFMGR_BO_LOCK_TYPE_ALWAYS;

	TBM_DBG("BUFMGR_LOCK_TYPE=%s\n", env ? env : "default:once");

	/* intialize bo_list */
	LIST_INITHEAD(&gBufMgr->bo_list);

	/* intialize surf_list */
	LIST_INITHEAD(&gBufMgr->surf_list);

	/* intialize surf_queue_list */
	LIST_INITHEAD(&gBufMgr->surf_queue_list);

	/* intialize debug_key_list */
	LIST_INITHEAD(&gBufMgr->debug_key_list);

#ifdef TBM_BUFMGR_INIT_TIME
	/* get the end tv */
	gettimeofday(&end_tv, NULL);
	TBM_INFO("tbm_bufmgr_init time: %ld ms", ((end_tv.tv_sec * 1000 + end_tv.tv_usec / 1000) - (start_tv.tv_sec * 1000 + start_tv.tv_usec / 1000)));
#endif

	pthread_mutex_unlock(&gLock);

	return gBufMgr;
}

tbm_bufmgr
tbm_bufmgr_init(int fd)
{
	tbm_bufmgr bufmgr;

	bufmgr = _tbm_bufmgr_init(fd, 0);

	return bufmgr;
}

void
tbm_bufmgr_deinit(tbm_bufmgr bufmgr)
{
	TBM_RETURN_IF_FAIL(TBM_BUFMGR_IS_VALID(bufmgr));

	_tbm_bufmgr_mutex_lock();
	pthread_mutex_lock(&gLock);
	_tbm_set_last_result(TBM_ERROR_NONE);

	if (!gBufMgr) {
		TBM_ERR("gBufmgr already destroy: bufmgr:%p\n", bufmgr);
		pthread_mutex_unlock(&gLock);
		_tbm_bufmgr_mutex_unlock();
		return;
	}

	bufmgr->ref_count--;
	if (bufmgr->ref_count > 0) {
		TBM_DBG("reduce a ref_count(%d) of tbm_bufmgr(%p)\n", bufmgr->ref_count, bufmgr);
		pthread_mutex_unlock(&gLock);
		_tbm_bufmgr_mutex_unlock();
		return;
	}

	/* destroy bo_list */
	if (!LIST_IS_EMPTY(&bufmgr->bo_list)) {
		tbm_bo bo = NULL, tmp;

		LIST_FOR_EACH_ENTRY_SAFE(bo, tmp, &bufmgr->bo_list, item_link) {
			TBM_ERR("Un-freed bo(%p, ref:%d)\n", bo, bo->ref_cnt);
			_tbm_bo_free(bo);
		}
		LIST_DELINIT(&bufmgr->bo_list);
	}

	/* destroy surf_list */
	if (!LIST_IS_EMPTY(&bufmgr->surf_list)) {
		tbm_surface_h surf = NULL, tmp;

		LIST_FOR_EACH_ENTRY_SAFE(surf, tmp, &bufmgr->surf_list, item_link) {
			TBM_ERR("Un-freed surf(%p, ref:%d)\n", surf, surf->refcnt);
			tbm_surface_destroy(surf);
		}
		LIST_DELINIT(&bufmgr->surf_list);
	}

	tbm_module_unload(bufmgr->module);

	// TODO: this is temporary. it will be removed after finishing refactoring the tbm_module.
	_tbm_bufmgr_reset_modlue_data(bufmgr);

	if (bufmgr->fd > 0)
		close(bufmgr->fd);

	TBM_INFO("destroy tbm_bufmgr(%p)\n", bufmgr);

	free(bufmgr);
	gBufMgr = NULL;

	pthread_mutex_unlock(&gLock);
	_tbm_bufmgr_mutex_unlock();
}

unsigned int
tbm_bufmgr_get_capability(tbm_bufmgr bufmgr)
{
	unsigned int capabilities = TBM_BUFMGR_CAPABILITY_NONE;

	_tbm_bufmgr_mutex_lock();
	_tbm_set_last_result(TBM_ERROR_NONE);

	TBM_BUFMGR_RETURN_VAL_IF_FAIL(TBM_BUFMGR_IS_VALID(bufmgr), TBM_BUFMGR_CAPABILITY_NONE);
	TBM_BUFMGR_RETURN_VAL_IF_FAIL(bufmgr == gBufMgr, TBM_BUFMGR_CAPABILITY_NONE);

	capabilities = bufmgr->capabilities;

	_tbm_bufmgr_mutex_unlock();

	return capabilities;
}

/* LCOV_EXCL_START */
char *
tbm_bufmgr_debug_tbm_info_get(tbm_bufmgr bufmgr)
{
	char app_name[255] = {0,}, title[512] = {0,};
	tbm_surface_debug_data *debug_old_data = NULL;
	char *str;
	int len = 1024*4;
	int c = 0;
	int size;
	tbm_error_e error;
	long pid = 0;

	pthread_mutex_lock(&gLock);
	_tbm_set_last_result(TBM_ERROR_NONE);

	if (!TBM_BUFMGR_IS_VALID(bufmgr) || (bufmgr != gBufMgr)) {
		TBM_ERR("invalid bufmgr\n");
		_tbm_set_last_result(TBM_ERROR_INVALID_PARAMETER);
		pthread_mutex_unlock(&gLock);
		return NULL;
	}

	str = malloc(len);
	if (!str) {
		TBM_ERR("Fail to allocate the string.\n");
		_tbm_set_last_result(TBM_ERROR_OUT_OF_MEMORY);
		pthread_mutex_unlock(&gLock);
		return NULL;
	}

	TBM_SNRPRINTF(str, len, c, "\n");
	pid = syscall(SYS_getpid);
	_tbm_util_get_appname_from_pid(pid, app_name);
	_tbm_util_get_appname_brief(app_name);
	TBM_SNRPRINTF(str, len, c, "===========================================TBM DEBUG: %s(%ld)===========================================\n",
		  app_name, pid);

	snprintf(title, 255, "%s", "no  surface     refcnt  width  height  bpp  size    n_b  n_p  flags    format    app_name              ");

	if (!LIST_IS_EMPTY(&bufmgr->debug_key_list)) {
		LIST_FOR_EACH_ENTRY(debug_old_data, &bufmgr->debug_key_list, item_link) {
			strncat(title, "  ", MAX_SIZE_N(title));
			strncat(title, debug_old_data->key, MAX_SIZE_N(title));
		}
	}

	TBM_SNRPRINTF(str, len, c, "[tbm_surface information]\n");
	TBM_SNRPRINTF(str, len, c, "%s\n", title);

	/* show the tbm_surface information in surf_list */
	if (!LIST_IS_EMPTY(&bufmgr->surf_list)) {
		tbm_surface_h surf = NULL;
		int surf_cnt = 0;

		LIST_FOR_EACH_ENTRY(surf, &bufmgr->surf_list, item_link) {
			char data[512] = {0,};
			unsigned int surf_pid = 0;
			int i;

			surf_pid = _tbm_surface_internal_get_debug_pid(surf);
			if (!surf_pid) {
				/* if pid is null, set the self_pid */
				surf_pid = syscall(SYS_getpid);;
			}

			memset(app_name, 0x0, 255 * sizeof(char));
			if (geteuid() == 0) {
				_tbm_util_get_appname_from_pid(surf_pid, app_name);
				_tbm_util_get_appname_brief(app_name);
			} else {
				snprintf(app_name, sizeof(app_name), "%d", surf_pid);
			}

			snprintf(data, 255, "%-3d %-11p   %-5d %-6u %-7u %-4u %-7u  %-3d  %-3d %-8d %-9s %-22s",
				  ++surf_cnt,
				  surf,
				  surf->refcnt,
				  surf->info.width,
				  surf->info.height,
				  surf->info.bpp,
				  surf->info.size / 1024,
				  surf->num_bos,
				  surf->num_planes,
				  surf->flags,
				  _tbm_surface_internal_format_to_str(surf->info.format) + 11,
				  app_name);

			if (!LIST_IS_EMPTY(&bufmgr->debug_key_list)) {
				LIST_FOR_EACH_ENTRY(debug_old_data, &bufmgr->debug_key_list, item_link) {
					char *value;

					strncat(data, "  ", MAX_SIZE_N(title));

					value = _tbm_surface_internal_get_debug_data(surf, debug_old_data->key);
					if (value)
						strncat(data, value, MAX_SIZE_N(title));
					else
						strncat(data, "none", MAX_SIZE_N(title));
				}
			}
			TBM_SNRPRINTF(str, len, c, "%s\n", data);

			for (i = 0; i < surf->num_bos; i++) {
				if (bufmgr->use_hal_tbm) {
					size = hal_tbm_bo_get_size((hal_tbm_bo *)surf->bos[i]->bo_data, (hal_tbm_error *)&error);
					if (error != TBM_ERROR_NONE)
						TBM_WRN("fail to get the size of bo.");
				} else if (bufmgr->backend_module_data) {
					size = bufmgr->bo_func->bo_get_size(surf->bos[i]->bo_data, &error);
					if (error != TBM_ERROR_NONE)
						TBM_WRN("fail to get the size of bo.");
				} else
					size = bufmgr->backend->bo_size(surf->bos[i]);
				TBM_SNRPRINTF(str, len, c, " bo:%-12p  %-26d%-10d\n",
					  surf->bos[i],
					  surf->bos[i]->ref_cnt,
					  size / 1024);
			}
		}
	} else
		TBM_SNRPRINTF(str, len, c, " no tbm_surfaces.\n");
	TBM_SNRPRINTF(str, len, c, "\n");

	TBM_SNRPRINTF(str, len, c, "[tbm_bo information]\n");
	TBM_SNRPRINTF(str, len, c, "no  bo          refcnt  size    lock_cnt  map_cnt  flags   surface     name\n");

	/* show the tbm_bo information in bo_list */
	if (!LIST_IS_EMPTY(&bufmgr->bo_list)) {
		int bo_cnt = 0;
		tbm_bo bo = NULL;
		tbm_key key = 0;

		LIST_FOR_EACH_ENTRY(bo, &bufmgr->bo_list, item_link) {
			if (bo->bufmgr->use_hal_tbm) {
				size = hal_tbm_bo_get_size((hal_tbm_bo *)bo->bo_data, (hal_tbm_error *)&error);
				if (error != TBM_ERROR_NONE)
					TBM_WRN("fail to get the size of bo.");
				key = (tbm_key)hal_tbm_bo_export_key((hal_tbm_bo *)bo->bo_data, (hal_tbm_error *)&error);
			} else if (bufmgr->backend_module_data) {
				size = bufmgr->bo_func->bo_get_size(bo->bo_data, &error);
				if (error != TBM_ERROR_NONE)
					TBM_WRN("fail to get the size of bo.");
				key = bufmgr->bo_func->bo_export_key(bo->bo_data, &error);
				if (error != TBM_ERROR_NONE)
					TBM_WRN("fail to get the tdm_key of bo.");
			} else {
				size = bufmgr->backend->bo_size(bo);
				key = bufmgr->backend->bo_export(bo);
			}
			TBM_SNRPRINTF(str, len, c, "%-3d %-11p   %-5d %-7d    %-6d    %-5u %-7d %-11p  %-4d\n",
				  ++bo_cnt,
				  bo,
				  bo->ref_cnt,
				  size / 1024,
				  bo->lock_cnt,
				  bo->map_cnt,
				  bo->flags,
				  bo->surface,
				  key);
		}
	} else
		TBM_SNRPRINTF(str, len, c, "no tbm_bos.\n");
	TBM_SNRPRINTF(str, len, c, "\n");

	TBM_SNRPRINTF(str, len, c, "========================================================================================================\n");

	pthread_mutex_unlock(&gLock);

	return str;
}

void
tbm_bufmgr_debug_show(tbm_bufmgr bufmgr)
{
	char * str;
	str = tbm_bufmgr_debug_tbm_info_get(bufmgr);
	if (str) {
		TBM_DBG("	%s", str);
		free(str);
	}
}

void
tbm_bufmgr_debug_trace(tbm_bufmgr bufmgr, int onoff)
{
	_tbm_bufmgr_mutex_lock();
	_tbm_set_last_result(TBM_ERROR_NONE);

	TBM_BUFMGR_RETURN_IF_FAIL(TBM_BUFMGR_IS_VALID(bufmgr));
	TBM_BUFMGR_RETURN_IF_FAIL(bufmgr == gBufMgr);

#ifdef TRACE
	TBM_LOG_D("bufmgr=%p onoff=%d\n", bufmgr, onoff);
	bTrace = onoff;
#endif

	_tbm_bufmgr_mutex_unlock();
}

void
tbm_bufmgr_debug_set_trace_mask(tbm_bufmgr bufmgr, tbm_bufmgr_debug_trace_mask mask, int set)
{
	_tbm_bufmgr_mutex_lock();
	_tbm_set_last_result(TBM_ERROR_NONE);

	TBM_BUFMGR_RETURN_IF_FAIL(TBM_BUFMGR_IS_VALID(bufmgr));
	TBM_BUFMGR_RETURN_IF_FAIL(bufmgr == gBufMgr);

	if (set == 1) {
		trace_mask |= mask;

		TBM_INFO("bufmgr=%p sets the trace_mask=%d\n", bufmgr, mask);
		if (trace_mask&TBM_BUFGMR_DEBUG_TRACE_BO)
			TBM_INFO("  TBM_BUFGMR_DEBUG_TRACE_BO");
		if (trace_mask&TBM_BUFGMR_DEBUG_TRACE_SURFACE_INTERNAL)
			TBM_INFO("  TBM_BUFGMR_DEBUG_TRACE_SURFACE_INTERNAL");
		if (trace_mask&TBM_BUFGMR_DEBUG_TRACE_SURFACE)
			TBM_INFO("  TBM_BUFGMR_DEBUG_TRACE_SURFACE");
		if (trace_mask&TBM_BUFGMR_DEBUG_TRACE_SURFACE_QUEUE)
			TBM_INFO("  TBM_BUFGMR_DEBUG_TRACE_SURFACE_QUEUE");
	} else if (set == 0) {
		trace_mask &= ~mask;

		TBM_INFO("bufmgr=%p unsets the trace_mask=%d\n", bufmgr, mask);
		if (trace_mask&TBM_BUFGMR_DEBUG_TRACE_BO)
			TBM_INFO("  TBM_BUFGMR_DEBUG_TRACE_BO");
		if (trace_mask&TBM_BUFGMR_DEBUG_TRACE_SURFACE_INTERNAL)
			TBM_INFO("  TBM_BUFGMR_DEBUG_TRACE_SURFACE_INTERNAL");
		if (trace_mask&TBM_BUFGMR_DEBUG_TRACE_SURFACE)
			TBM_INFO("  TBM_BUFGMR_DEBUG_TRACE_SURFACE");
		if (trace_mask&TBM_BUFGMR_DEBUG_TRACE_SURFACE_QUEUE)
			TBM_INFO("  TBM_BUFGMR_DEBUG_TRACE_SURFACE_QUEUE");
	} else {
		TBM_WRN("set value is wrong.(set=%d)", set);
	}

	_tbm_bufmgr_mutex_unlock();
}

void
tbm_bufmgr_debug_dump_set_scale(double scale)
{
	pthread_mutex_lock(&gLock);
	_tbm_set_last_result(TBM_ERROR_NONE);
	scale_factor = scale;
	pthread_mutex_unlock(&gLock);
}

int
tbm_bufmgr_debug_get_ref_count(void)
{
	int refcnt;

	pthread_mutex_lock(&gLock);

	_tbm_set_last_result(TBM_ERROR_NONE);

	refcnt = (gBufMgr) ? gBufMgr->ref_count : 0;

	pthread_mutex_unlock(&gLock);

	return refcnt;
}

int
tbm_bufmgr_debug_queue_dump(char *path, int count, int onoff)
{
	pthread_mutex_lock(&gLock);
	_tbm_set_last_result(TBM_ERROR_NONE);

	if (onoff == 0) {
		TBM_DBG("count=%d onoff=%d\n", count, onoff);
		b_dump_queue = 0;
		tbm_surface_internal_dump_end();
	} else {
		int w, h;

		if (path == NULL) {
			TBM_ERR("path is null");
			pthread_mutex_unlock(&gLock);
			return 0;
		}
		TBM_DBG("path=%s count=%d onoff=%d\n", path, count, onoff);

		if (_tbm_util_get_max_surface_size(&w, &h) == 0) {
			TBM_ERR("Fail to get tbm_surface size.\n");
			pthread_mutex_unlock(&gLock);
			return 0;
		}

		tbm_surface_internal_dump_with_scale_start(path, w, h, count, scale_factor);
		scale_factor = 0;

		b_dump_queue = 1;
	}

	pthread_mutex_unlock(&gLock);

	return 1;
}

int
tbm_bufmgr_debug_dump_all(char *path)
{
	int w, h, count = 0;
	tbm_surface_h surface = NULL;

	pthread_mutex_lock(&gLock);
	_tbm_set_last_result(TBM_ERROR_NONE);

	if (!path) {
		TBM_ERR("path is null.\n");
		pthread_mutex_unlock(&gLock);
		return 0;
	}

	TBM_DBG("path=%s\n", path);

	count = _tbm_util_get_max_surface_size(&w, &h);
	if (count == 0) {
		TBM_ERR("No tbm_surface.\n");
		pthread_mutex_unlock(&gLock);
		return 0;
	}

	tbm_surface_internal_dump_with_scale_start(path, w, h, count, scale_factor);
	scale_factor = 0;

	LIST_FOR_EACH_ENTRY(surface, &gBufMgr->surf_list, item_link)
		tbm_surface_internal_dump_buffer(surface, "dump_all");

	tbm_surface_internal_dump_end();

	pthread_mutex_unlock(&gLock);

	return 1;
}

/* internal function */
tbm_bufmgr
_tbm_bufmgr_get_bufmgr(void)
{
	return gBufMgr;
}

int
tbm_bufmgr_bind_native_display(tbm_bufmgr bufmgr, void *native_display)
{
	tbm_error_e error;

	_tbm_bufmgr_mutex_lock();
	_tbm_set_last_result(TBM_ERROR_NONE);

	TBM_BUFMGR_RETURN_VAL_IF_FAIL(TBM_BUFMGR_IS_VALID(bufmgr), 0);

	error = tbm_module_bufmgr_bind_native_display(bufmgr->module, native_display);
	if (error != TBM_ERROR_NONE) {
		_tbm_set_last_result(error);
		_tbm_bufmgr_mutex_unlock();

		if (error == TBM_ERROR_NOT_SUPPORTED) {
			TBM_WRN("Not supported, so skip: tbm_bufmgr(%p) native_display(%p)", bufmgr, native_display);
			return 1;
		}

		return 0;
	}

	TBM_INFO("tbm_bufmgr(%p) native_display(%p)\n", bufmgr, native_display);

	_tbm_bufmgr_mutex_unlock();

	return 1;
}

tbm_bufmgr
tbm_bufmgr_server_init(void)
{
	tbm_bufmgr bufmgr;

	bufmgr = _tbm_bufmgr_init(-1, 1);

	return bufmgr;
}

int
tbm_bufmgr_set_bo_lock_type(tbm_bufmgr bufmgr, tbm_bufmgr_bo_lock_type bo_lock_type)
{
	_tbm_bufmgr_mutex_lock();
	_tbm_set_last_result(TBM_ERROR_NONE);

	TBM_BUFMGR_RETURN_VAL_IF_FAIL(TBM_BUFMGR_IS_VALID(gBufMgr), 0);
	TBM_BUFMGR_RETURN_VAL_IF_FAIL(bufmgr == gBufMgr, 0);

	pthread_mutex_lock(&gLock);
	gBufMgr->bo_lock_type = bo_lock_type;
	pthread_mutex_unlock(&gLock);

	TBM_INFO("The bo_lock_type of the bo is %d\n", bo_lock_type);

	_tbm_bufmgr_mutex_unlock();

	return 1;
}


int tbm_bufmgr_get_fd_limit(void)
{
	struct rlimit lim;

	if (getrlimit(RLIMIT_NOFILE, &lim))
		return 1024;

	return (int)lim.rlim_cur;
}

tbm_bufmgr tbm_bufmgr_get(void)
{
	return gBufMgr;
}


tbm_bo
tbm_bufmgr_internal_find_bo(tbm_bufmgr bufmgr, tbm_bo bo)
{
	tbm_bo bo2 = NULL;

	TBM_RETURN_VAL_IF_FAIL(TBM_BUFMGR_IS_VALID(gBufMgr), NULL);
	TBM_RETURN_VAL_IF_FAIL(bufmgr == gBufMgr, NULL);

	if (LIST_IS_EMPTY(&bufmgr->bo_list))
		return NULL;

	LIST_FOR_EACH_ENTRY(bo2, &bufmgr->bo_list, item_link) {
		if (bo2->bo_data == bo->bo_data) {
			return bo2;
		}
	}

	return NULL;
}

tbm_bo
tbm_bufmgr_internal_alloc_bo(tbm_bufmgr bufmgr, int size, int flags, tbm_error_e *error)
{
	tbm_bo bo;
	tbm_backend_bo_data *bo_data;

	_tbm_bufmgr_check_bo_cnt(bufmgr);

	bo = calloc(1, sizeof(struct _tbm_bo));
	if (!bo) {
		/* LCOV_EXCL_START */
		TBM_ERR("memory allocationc failed.");
		*error = TBM_ERROR_OUT_OF_MEMORY;
		return NULL;
		/* LCOV_EXCL_STOP */
	}

	bo_data = tbm_module_bufmgr_bo_alloc(bufmgr->module, bo, size, flags, error);
	if (!bo_data) {
		/* LCOV_EXCL_START */
		TBM_ERR("tbm_module_bufmgr_bo_alloc failed. size:%d flags:%s error:%d", size, _tbm_flag_to_str(flags), *error);
		free(bo);
		return NULL;
		/* LCOV_EXCL_STOP */
	}
	bo->bo_data = bo_data;
	bo->priv = (void *)bo_data; // TODO: this will be DEPRECATED.

	_tbm_bufmgr_initialize_bo(bufmgr, bo, flags);

	return bo;
}


/* LCOV_EXCL_START */

tbm_bo
tbm_bufmgr_internal_alloc_bo_with_format(tbm_bufmgr bufmgr, int format, int bo_idx, int width,
						int height, int bpp, tbm_bo_memory_type flags, tbm_error_e *error)
{
	tbm_bo bo = NULL;

	_tbm_bufmgr_mutex_lock();
	_tbm_set_last_result(TBM_ERROR_NONE);

	TBM_BUFMGR_RETURN_VAL_IF_FAIL(TBM_BUFMGR_IS_VALID(bufmgr), NULL);

	_tbm_bufmgr_check_bo_cnt(bufmgr);

	bo = calloc(1, sizeof(struct _tbm_bo));
	if (!bo) {
		/* LCOV_EXCL_START */
		TBM_ERR("memory allocationc failed.");
		_tbm_set_last_result(TBM_ERROR_OUT_OF_MEMORY);
		/* LCOV_EXCL_STOP */
		goto fail;
	}

	bo->bo_data = tbm_module_bufmgr_bo_alloc_with_format(bufmgr->module, format, bo_idx, width, height, bpp, flags, error);
	if (!bo->bo_data) {
		/* LCOV_EXCL_START */
		_tbm_set_last_result(*error);
		/* LCOV_EXCL_STOP */
		goto fail;
	}

	_tbm_bufmgr_initialize_bo(bufmgr, bo, flags);

	_tbm_bufmgr_mutex_unlock();

	return bo;

fail:
	if (bo)
		free(bo);
	_tbm_bufmgr_mutex_unlock();

	return NULL;
}

tbm_bo
tbm_bufmgr_internal_alloc_bo_with_bo_data(tbm_bufmgr bufmgr, tbm_backend_bo_data *bo_data, int flags)
{
	tbm_bo bo, bo2 = NULL;

	_tbm_bufmgr_mutex_lock();
	_tbm_set_last_result(TBM_ERROR_NONE);

	TBM_BUFMGR_RETURN_VAL_IF_FAIL(TBM_BUFMGR_IS_VALID(bufmgr), NULL);
	TBM_BUFMGR_RETURN_VAL_IF_FAIL(bo_data, NULL);

	_tbm_bufmgr_check_bo_cnt(bufmgr);

	bo = calloc(1, sizeof(struct _tbm_bo));
	if (!bo) {
		/* LCOV_EXCL_START */
		TBM_ERR("memory allocationc failed.");
		_tbm_set_last_result(TBM_ERROR_OUT_OF_MEMORY);
		_tbm_bufmgr_mutex_unlock();
		return NULL;
		/* LCOV_EXCL_STOP */
	}
	bo->bo_data = bo_data;
	bo->get_from_hal_surface = 1;

	// return an existed bo if the bo is already created with the same bo_data.
	bo2 = tbm_bufmgr_internal_find_bo(bufmgr, bo);
	if (bo2) {
		bo2->ref_cnt++;
		free(bo);
		_tbm_bufmgr_mutex_unlock();
		return bo2;
	}

	_tbm_bufmgr_initialize_bo(bufmgr, bo, flags);

	TBM_TRACE_BO("bo(%p) refcnt(%d), flag(%s)", bo, bo->ref_cnt, _tbm_flag_to_str(bo->flags));

	_tbm_bufmgr_mutex_unlock();

	return bo;
}

tbm_bo
tbm_bufmgr_internal_import_bo_with_key(tbm_bufmgr bufmgr, unsigned int key, tbm_error_e *error)
{
	tbm_bo bo, bo2;
	int flags;

	_tbm_bufmgr_check_bo_cnt(bufmgr);

	bo = calloc(1, sizeof(struct _tbm_bo));
	if (!bo) {
		/* LCOV_EXCL_START */
		TBM_ERR("memory allocationc failed.");
		*error = TBM_ERROR_OUT_OF_MEMORY;
		return NULL;
		/* LCOV_EXCL_STOP */
	}

	bo->bo_data = tbm_module_bufmgr_bo_import_key(bufmgr->module, bo, key, error);
	if (!bo->bo_data) {
		/* LCOV_EXCL_START */
		TBM_ERR("tbm_module_bufmgr_bo_import_key failed. tbm_key:%d", key);
		free(bo);
		return NULL;
		/* LCOV_EXCL_STOP */
	}

	// return the existed bo2 if bo->bo_data and bo2->bo_data is the same
	bo2 = tbm_bufmgr_internal_find_bo(bufmgr, bo);
	if (bo2) {
		TBM_TRACE_BO("find bo(%p) ref(%d) key(%d) flag(%s) in list",
					bo2, bo2->ref_cnt, key, _tbm_flag_to_str(bo2->flags));
		bo2->ref_cnt++;
		free(bo);
		return bo2;
	}

	// TODO: refactoring tbm_module_bo
	if (bufmgr->use_hal_tbm) {
		flags = (tbm_bo_memory_type)hal_tbm_bo_get_memory_types((hal_tbm_bo *)bo->bo_data, (hal_tbm_error *)error);
		if (*error != TBM_ERROR_NONE) {
			TBM_ERR("fail to get the bo flags(memory_types)");
			flags = TBM_BO_DEFAULT;
		}
	} else if (bufmgr->backend_module_data) {
		flags = bufmgr->bo_func->bo_get_memory_types(bo->bo_data, error);
		if (error != TBM_ERROR_NONE) {
			TBM_ERR("fail to get the bo flags(memory_types)");
			flags = TBM_BO_DEFAULT;
		}
	} else {
		if (bufmgr->backend->bo_get_flags)
			flags = bufmgr->backend->bo_get_flags(bo);
		else
			flags = TBM_BO_DEFAULT;
	}

	_tbm_bufmgr_initialize_bo(bufmgr, bo, flags);

	return bo;
}

tbm_bo
tbm_bufmgr_internal_import_bo_with_fd(tbm_bufmgr bufmgr, tbm_fd fd, tbm_error_e *error)
{
	tbm_bo bo, bo2;
	int flags;

	_tbm_bufmgr_check_bo_cnt(bufmgr);

	bo = calloc(1, sizeof(struct _tbm_bo));
	if (!bo) {
		/* LCOV_EXCL_START */
		TBM_ERR("memory allocationc failed.");
		*error = TBM_ERROR_OUT_OF_MEMORY;
		return NULL;
		/* LCOV_EXCL_STOP */
	}

	bo->bo_data = tbm_module_bufmgr_bo_import_fd(bufmgr->module, bo, fd, error);
	if (!bo->bo_data) {
		/* LCOV_EXCL_START */
		TBM_ERR("tbm_module_bufmgr_bo_import_fd failed. tbm_fd:%d", fd);
		free(bo);
		return NULL;
		/* LCOV_EXCL_STOP */
	}

	// return the existed bo2 if bo->bo_data and bo2->bo_data is the same
	bo2 = tbm_bufmgr_internal_find_bo(bufmgr, bo);
	if (bo2) {
		TBM_TRACE_BO("find bo(%p) ref(%d) fd(%d) flag(%s) in list",
					bo2, bo2->ref_cnt, fd, _tbm_flag_to_str(bo2->flags));
		bo2->ref_cnt++;
		free(bo);
		return bo2;
	}

	// TODO: refactoring tbm_module_bo
	if (bufmgr->use_hal_tbm) {
		flags = (tbm_bo_memory_type)hal_tbm_bo_get_memory_types((hal_tbm_bo *)bo->bo_data, (hal_tbm_error *)error);
		if (error != TBM_ERROR_NONE) {
			TBM_ERR("fail to get the bo flags(memory_types)");
			flags = TBM_BO_DEFAULT;
		}
	} else if (bufmgr->backend_module_data) {
		flags = bufmgr->bo_func->bo_get_memory_types(bo->bo_data, error);
		if (error != TBM_ERROR_NONE) {
			TBM_ERR("fail to get the bo flags(memory_types)");
			flags = TBM_BO_DEFAULT;
		}
	} else {
		if (bufmgr->backend->bo_get_flags)
			flags = bufmgr->backend->bo_get_flags(bo);
		else
			flags = TBM_BO_DEFAULT;
	}

	_tbm_bufmgr_initialize_bo(bufmgr, bo, flags);

	return bo;
}

/* LCOV_EXCL_STOP */
