/**************************************************************************

libtbm

Copyright 2012 Samsung Electronics co., Ltd. All Rights Reserved.

Contact: SooChan Lim <sc1.lim@samsung.com>, Sangjin Lee <lsj119@samsung.com>
Boram Park <boram1288.park@samsung.com>, Changyeon Lee <cyeon.lee@samsung.com>

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sub license, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
IN NO EVENT SHALL PRECISION INSIGHT AND/OR ITS SUPPLIERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

**************************************************************************/

#define WL_HIDE_DEPRECATED

#include "config.h"

#include <xf86drm.h>
#include <grp.h>

#include "tbm_bufmgr_int.h"

#include "wayland-tbm-drm-auth-server-protocol.h"

struct wayland_tbm_drm_auth_server {
	struct wl_display *display;
	struct wl_global *wl_tbm_drm_auth_global;

	char *device_name;
	uint32_t fd;
	uint32_t flags;
};

#define MIN(x, y) (((x) < (y)) ? (x) : (y))

struct wayland_tbm_drm_auth_server *tbm_drm_auth_srv;

static int tbm_drm_master_fd = -1;

/* LCOV_EXCL_START */
static void
_send_server_auth_info(struct wayland_tbm_drm_auth_server *tbm_drm_auth_srv,
		       struct wl_resource *resource)
{
	int fd = -1;
	uint32_t capabilities;
	char *device_name = NULL;
	drm_magic_t magic = 0;

	fd = open(tbm_drm_auth_srv->device_name, O_RDWR | O_CLOEXEC);
	if (fd == -1 && errno == EINVAL) {
		fd = open(tbm_drm_auth_srv->device_name, O_RDWR);
		if (fd != -1) {
			if (fcntl(fd, F_SETFD, fcntl(fd, F_GETFD) | FD_CLOEXEC) == -1) {
				TBM_ERR("failed to set fd\n");
				goto fini;
			}
		}
	}

	if (fd < 0) {
		TBM_ERR("failed to open drm : device_name, %s\n", tbm_drm_auth_srv->device_name);

		wl_resource_post_error(resource, WL_TBM_DRM_AUTH_ERROR_AUTHENTICATE_FAIL,
				       "authenicate failed::open_drm");
		goto fini;
	}

	if (drmGetMagic(fd, &magic) < 0) {
		if (errno != EACCES) {
			TBM_ERR("failed to get magic\n");

			wl_resource_post_error(resource, WL_TBM_DRM_AUTH_ERROR_AUTHENTICATE_FAIL,
					       "authenicate failed::get_magic");
			goto fini;
		}
	}

	if (drmAuthMagic(tbm_drm_auth_srv->fd, magic) < 0) {
		TBM_ERR("failed to authenticate magic\n");

		wl_resource_post_error(resource, WL_TBM_DRM_AUTH_ERROR_AUTHENTICATE_FAIL,
				       "authenicate failed::auth_magic");
		goto fini;
	}

	capabilities = tbm_drm_auth_srv->flags;
	device_name = tbm_drm_auth_srv->device_name;

	/* send */
	wl_tbm_drm_auth_send_authentication_info(resource, device_name, capabilities, fd);

fini:
	if (fd >= 0)
		close(fd);

	if (device_name && device_name != tbm_drm_auth_srv->device_name)
		free(device_name);

}

static void
_wayland_tbm_drm_auth_server_impl_get_authentication_info(struct wl_client *client,
		struct wl_resource *resource)
{
	struct wayland_tbm_drm_auth_server *tbm_drm_auth_srv = wl_resource_get_user_data(resource);

	/* if display server is the client of the host display server, for embedded server */
	_send_server_auth_info(tbm_drm_auth_srv, resource);
}


static const struct wl_tbm_drm_auth_interface _wayland_tbm_drm_auth_server_implementation = {
	_wayland_tbm_drm_auth_server_impl_get_authentication_info,
};

static void
_wayland_tbm_drm_auth_server_bind_cb(struct wl_client *client, void *data,
			    uint32_t version,
			    uint32_t id)
{
	struct wl_resource *resource;

	resource = wl_resource_create(client, &wl_tbm_drm_auth_interface, MIN(version, 1), id);
	if (!resource) {
		wl_client_post_no_memory(client);
		return;
	}

	wl_resource_set_implementation(resource,
				       &_wayland_tbm_drm_auth_server_implementation,
				       data,
				       NULL);
}

static int
_tbm_getgrnam_r(const char *name)
{
	struct group *grp = NULL;
	struct group *grp_res = NULL;
	char* buf = NULL;
	size_t buf_len;
	int ret;
	int id;

	buf_len = sysconf(_SC_GETGR_R_SIZE_MAX);
	if (buf_len == -1)
		buf_len = 2048;

	buf = calloc(1, buf_len * sizeof(char));
	if (!buf) {
		TBM_ERR("creating buffer failed\n");
		goto failed;
	}

	grp = calloc(1, sizeof(struct group));
	if (!grp) {
		TBM_ERR("creating group failed\n");
		goto failed;
	}

	ret = getgrnam_r(name, grp, buf, buf_len, &grp_res);
	if (ret < 0) {
		TBM_ERR("getgrnam_r failed errno:%d(%m)\n", ret);
		goto failed;
	}

	if (grp_res == NULL) {
		TBM_ERR("finding name:%s group failed\n", name);
		goto failed;
	}

	id = grp->gr_gid;
	free(buf);
	free(grp);

	return id;

failed:
	if (buf)
		free(buf);
	if (grp)
		free(grp);

	return -1;
}

static void
_tbm_drm_auth_socket_init(struct wayland_tbm_drm_auth_server *tbm_drm_auth_srv)
{
	const char *dir = NULL;
	char socket_path[128];
	int ret = -1;
	uid_t uid;
	gid_t gid;

	dir = getenv("XDG_RUNTIME_DIR");
	if (!dir) {
		TBM_WRN("getting XDG_RUNTIME_DIR failed\n");
		return;
	}

	snprintf(socket_path, sizeof(socket_path), "%s/%s", dir, "tbm-drm-auth");

	ret = chmod(socket_path, 509);
	if (ret < 0) {
		TBM_WRN("changing modes of socket file failed:%s (%m)\n", socket_path);
		return;
	}

	ret = _tbm_getgrnam_r("root");
	if (ret < 0) {
		TBM_WRN("getting uid failed\n");
		return;
	}
	uid = ret;

	ret = _tbm_getgrnam_r("display");
	if (ret < 0) {
		TBM_WRN("getting gid failed\n");
		return;
	}
	gid = ret;

	ret = chown(socket_path, uid, gid);
	if (ret < 0) {
		TBM_WRN("changing owner of socket file failed:%s (%m)\n", socket_path);
		return;
	}
}

int
tbm_drm_helper_wl_auth_server_init(void *wl_display, int fd, const char *device_name, uint32_t flags)
{
	if (!tbm_drm_auth_srv) {
		TBM_RETURN_VAL_IF_FAIL(wl_display != NULL, 0);

		tbm_drm_auth_srv = calloc(1, sizeof(struct wayland_tbm_drm_auth_server));
		TBM_RETURN_VAL_IF_FAIL(tbm_drm_auth_srv != NULL, 0);

		tbm_drm_auth_srv->display = (struct wl_display *)wl_display;
		if (!device_name) {
			tbm_drm_auth_srv->device_name = drmGetDeviceNameFromFd(fd);
			if (!tbm_drm_auth_srv->device_name) {
					TBM_ERR("fail to get device name!\n");
					return 0;
			}
		} else {
			tbm_drm_auth_srv->device_name = strdup(device_name);
		}
		tbm_drm_auth_srv->fd = fd;
		tbm_drm_auth_srv->flags = flags;

		if (wl_display_add_socket(tbm_drm_auth_srv->display, "tbm-drm-auth")) {
			TBM_ERR("[TBM_DRM] fail to add socket\n");

			if (tbm_drm_auth_srv->device_name)
				free(tbm_drm_auth_srv->device_name);

			free(tbm_drm_auth_srv);
			tbm_drm_auth_srv = NULL;

			return 0;
		}

		_tbm_drm_auth_socket_init(tbm_drm_auth_srv);

		/* init the client resource list */
		tbm_drm_auth_srv->wl_tbm_drm_auth_global = wl_global_create(tbm_drm_auth_srv->display, &wl_tbm_drm_auth_interface, 1,
					 tbm_drm_auth_srv, _wayland_tbm_drm_auth_server_bind_cb);
	}

	return 1;
}

void
tbm_drm_helper_wl_auth_server_deinit(void)
{
	if (tbm_drm_auth_srv) {
		wl_global_destroy(tbm_drm_auth_srv->wl_tbm_drm_auth_global);

		if (tbm_drm_auth_srv->device_name)
			free(tbm_drm_auth_srv->device_name);

		free(tbm_drm_auth_srv);
		tbm_drm_auth_srv = NULL;
	}
}

int
tbm_drm_helper_get_master_fd(void)
{
	int new_fd, flags, fd = -1;

	fd = tbm_drm_master_fd;
	if (fd == -1) {
		TBM_INFO("no presetted TBM DRM MASTER FD");
		return -1;
	}

	TBM_INFO("TBM DRM MASTER FD: %d\n", fd);

	flags = fcntl(fd, F_GETFD);
	if (flags == -1) {
		TBM_ERR("fcntl failed: %m");
		return -1;
	}

	new_fd = dup(fd);
	if (new_fd < 0) {
		TBM_ERR("dup failed: %m");
		return -1;
	}

	if (fcntl(new_fd, F_SETFD, flags|FD_CLOEXEC) == -1) {
		TBM_ERR("failed to set fd\n");
		close(new_fd);
		return -1;
	}

	TBM_INFO("Return MASTER_FD: %d\n", new_fd);

	return new_fd;
}

void
tbm_drm_helper_set_tbm_master_fd(int fd)
{
	int fd_max = tbm_bufmgr_get_fd_limit();

	if (tbm_drm_master_fd == fd)
		return;

	if (fd < 0 || fd > fd_max) {
		TBM_ERR("%d out of fd range\n", fd);
		return;
	}

	if (tbm_drm_master_fd != -1)
		TBM_WRN("already has TBM DRM MASTER FD: %d\n", tbm_drm_master_fd);

	tbm_drm_master_fd = fd;

	TBM_INFO("TBM DRM MASTER FD: %d\n", tbm_drm_master_fd);
}

void
tbm_drm_helper_unset_tbm_master_fd(void)
{
	tbm_drm_master_fd = -1;
	TBM_INFO("TBM DRM MASTER FD: %d\n", tbm_drm_master_fd);
}
/* LCOV_EXCL_STOP */

